"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
var assert_1 = require("assert");
var Path = require("path");
var binary = process.__nexe;
assert_1.ok(binary);
var manifest = binary.resources;
var directories = {};
var isString = function (x) { return typeof x === 'string' || x instanceof String; };
var isNotFile = function () { return false; };
var isNotDirectory = isNotFile;
var isFile = function () { return true; };
var isDirectory = isFile;
var fs = require('fs');
var originalReadFile = fs.readFile;
var originalReadFileSync = fs.readFileSync;
var originalCreateReadStream = fs.createReadStream;
var originalReaddir = fs.readdir;
var originalReaddirSync = fs.readdirSync;
var originalStatSync = fs.statSync;
var originalStat = fs.stat;
var originalRealpath = fs.realpath;
var originalRealpathSync = fs.realpathSync;
var resourceStart = binary.layout.resourceStart;
var statTime = function () {
    var stat = binary.layout.stat;
    return {
        dev: 0,
        ino: 0,
        nlink: 0,
        rdev: 0,
        uid: 123,
        gid: 500,
        blksize: 4096,
        blocks: 0,
        atime: new Date(stat.atime),
        atimeMs: stat.atime.getTime(),
        mtime: new Date(stat.mtime),
        mtimeMs: stat.mtime.getTime(),
        ctime: new Date(stat.ctime),
        ctimMs: stat.ctime.getTime(),
        birthtime: new Date(stat.birthtime),
        birthtimeMs: stat.birthtime.getTime()
    };
};
var createStat = function (directoryExtensions, fileExtensions) {
    if (!fileExtensions) {
        return Object.assign({}, binary.layout.stat, directoryExtensions, { size: 0 }, statTime());
    }
    var size = directoryExtensions[1];
    return Object.assign({}, binary.layout.stat, fileExtensions, { size: size }, statTime());
};
var ownStat = function (path) {
    var key = Path.resolve(path);
    if (directories[key]) {
        return createStat({ isDirectory: isDirectory, isFile: isNotFile });
    }
    if (manifest[key]) {
        return createStat(manifest[key], { isFile: isFile, isDirectory: isNotDirectory });
    }
};
function makeLong(filepath) {
    return Path._makeLong && Path._makeLong(filepath);
}
var setupManifest = function () {
    Object.keys(manifest).forEach(function (key) {
        var entry = manifest[key];
        var absolutePath = Path.resolve(key);
        var longPath = makeLong(absolutePath);
        var normalizedPath = Path.normalize(key);
        if (!manifest[absolutePath]) {
            manifest[absolutePath] = entry;
        }
        if (longPath && !manifest[longPath]) {
            manifest[longPath] = entry;
        }
        if (!manifest[normalizedPath]) {
            manifest[normalizedPath] = manifest[key];
        }
        var currentDir = Path.dirname(absolutePath);
        var prevDir = absolutePath;
        while (currentDir !== prevDir) {
            directories[currentDir] = directories[currentDir] || {};
            directories[currentDir][Path.basename(prevDir)] = true;
            var longDir = makeLong(currentDir);
            if (longDir && !directories[longDir]) {
                directories[longDir] = directories[currentDir];
            }
            prevDir = currentDir;
            currentDir = Path.dirname(currentDir);
        }
    });
    setupManifest = function () { };
};
//naive patches intended to work for most use cases
var nfs = {
    realpath: function realpath(path, options, cb) {
        setupManifest();
        if (isString(path) && manifest[path]) {
            return process.nextTick(function () { return cb(null, path); });
        }
        return originalRealpath.call(fs, path, options, cb);
    },
    realpathSync: function realpathSync(path, options) {
        setupManifest();
        if (isString(path) && manifest[path]) {
            return path;
        }
        return originalRealpathSync.call(fs, path, options);
    },
    readdir: function readdir(path, options, callback) {
        setupManifest();
        path = path.toString();
        if ('function' === typeof options) {
            callback = options;
            options = { encoding: 'utf8' };
        }
        var dir = directories[Path.resolve(path)];
        if (dir) {
            process.nextTick(function () {
                //todo merge with original?
                callback(null, Object.keys(dir));
            });
        }
        else {
            return originalReaddir.apply(fs, arguments);
        }
    },
    readdirSync: function readdirSync(path, options) {
        setupManifest();
        path = path.toString();
        var dir = directories[Path.resolve(path)];
        if (dir) {
            return Object.keys(dir);
        }
        return originalReaddirSync.apply(fs, arguments);
    },
    readFile: function readFile(file, options, callback) {
        setupManifest();
        var entry = manifest[file] || manifest[Path.resolve(file)];
        if (!entry || !isString(file)) {
            return originalReadFile.apply(fs, arguments);
        }
        var offset = entry[0], length = entry[1];
        var resourceOffset = resourceStart + offset;
        var encoding = isString(options) ? options : null;
        callback = typeof options === 'function' ? options : callback;
        fs.open(process.execPath, 'r', function (err, fd) {
            if (err)
                return callback(err, null);
            fs.read(fd, Buffer.alloc(length), 0, length, resourceOffset, function (error, bytesRead, result) {
                if (error) {
                    return fs.close(fd, function () {
                        callback(error, null);
                    });
                }
                fs.close(fd, function (err) {
                    if (err) {
                        return callback(err, result);
                    }
                    callback(err, encoding ? result.toString(encoding) : result);
                });
            });
        });
    },
    createReadStream: function createReadStream(file, options) {
        setupManifest();
        var entry = manifest[file] || manifest[Path.resolve(file)];
        if (!entry || !isString(file)) {
            return originalCreateReadStream.apply(fs, arguments);
        }
        var offset = entry[0], length = entry[1];
        var resourceOffset = resourceStart + offset;
        var opts = !options ? {} : isString(options) ? { encoding: options } : options;
        return fs.createReadStream(process.execPath, Object.assign({}, opts, {
            start: resourceOffset,
            end: resourceOffset + length
        }));
    },
    readFileSync: function readFileSync(file, options) {
        setupManifest();
        var entry = manifest[file] || manifest[Path.resolve(file)];
        console.log('reading', file, 'entry', entry);
        if (!entry || !isString(file)) {
            console.log('no entry');
            return originalReadFileSync.apply(fs, arguments);
        }
        var offset = entry[0], length = entry[1];
        var resourceOffset = resourceStart + offset;
        var encoding = isString(options) ? options : null;
        var fd = fs.openSync(process.execPath, 'r');
        var result = Buffer.alloc(length);
        fs.readSync(fd, result, 0, length, resourceOffset);
        fs.closeSync(fd);
        return encoding ? result.toString(encoding) : result;
    },
    statSync: function statSync(path) {
        var stat = isString(path) && ownStat(path);
        if (stat) {
            return stat;
        }
        return originalStatSync.apply(fs, arguments);
    },
    stat: function stat(path, callback) {
        var stat = isString(path) && ownStat(path);
        if (stat) {
            process.nextTick(function () {
                callback(null, stat);
            });
        }
        else {
            return originalStat.apply(fs, arguments);
        }
    }
};
var patches = process.nexe.patches;
delete process.nexe;
patches.internalModuleReadFile = function (original) {
    var args = [];
    for (var _i = 1; _i < arguments.length; _i++) {
        args[_i - 1] = arguments[_i];
    }
    var filepath = args[0];
    setupManifest();
    if (manifest[filepath]) {
        return nfs.readFileSync(filepath, 'utf-8');
    }
    return original.call.apply(original, [this].concat(args));
};
patches.internalModuleStat = function (original) {
    var args = [];
    for (var _i = 1; _i < arguments.length; _i++) {
        args[_i - 1] = arguments[_i];
    }
    setupManifest();
    var filepath = args[0];
    if (manifest[filepath]) {
        return 0;
    }
    if (directories[filepath]) {
        return 1;
    }
    return original.call.apply(original, [this].concat(args));
};
Object.assign(fs, nfs);
